from __future__ import absolute_import

import os
import getpass
from glob import glob
import sys

try:
    raw_input
except NameError:
    raw_input = input

try:
    from collections import OrderedDict
except ImportError:
    from .packages.ordereddict import OrderedDict  # NOQA

__all__ = ('Args', )

STDOUT = sys.stdout.write
NEWLINES = ('\n', '\r', '\r\n')


class Args(object):
    """CLI Argument management."""

    def __init__(self, args=None, no_argv=False):
        if args is None:
            if not no_argv:
                self._args = sys.argv[1:]
            else:
                self._args = []
        else:
            self._args = args

    def __len__(self):
        return len(self._args)

    def __repr__(self):
        return '<args %s>' % (repr(self._args))

    def __getitem__(self, i):
        try:
            return self.all[i]
        except IndexError:
            return None

    def __contains__(self, x):
        return self.first(x) is not None

    def get(self, x):
        """Returns argument at given index, else none."""
        try:
            return self.all[x]
        except IndexError:
            return None

    def get_with(self, x):
        """Returns first argument that contains given string."""
        return self.all[self.first_with(x)]

    def remove(self, x):
        """Removes given arg (or list thereof) from Args object."""

        def _remove(x):
            found = self.first(x)
            if found is not None:
                self._args.pop(found)

        if is_collection(x):
            for item in x:
                _remove(x)
        else:
            _remove(x)

    def pop(self, x):
        """Removes and Returns value at given index, else none."""
        try:
            return self._args.pop(x)
        except IndexError:
            return None

    def any_contain(self, x):
        """Tests if given string is contained in any stored argument."""

        return bool(self.first_with(x))

    def contains(self, x):
        """Tests if given object is in arguments list.
           Accepts strings and lists of strings."""

        return self.__contains__(x)

    def first(self, x):
        """Returns first found index of given value (or list of values)"""

        def _find(x):
            try:
                return self.all.index(str(x))
            except ValueError:
                return None

        if is_collection(x):
            for item in x:
                found = _find(item)
                if found is not None:
                    return found
            return None
        else:
            return _find(x)

    def first_with(self, x):
        """Returns first found index containing value (or list of values)"""

        def _find(x):
            try:
                for arg in self.all:
                    if x in arg:
                        return self.all.index(arg)
            except ValueError:
                return None

        if is_collection(x):
            for item in x:
                found = _find(item)
                if found:
                    return found
            return None
        else:
            return _find(x)

    def first_without(self, x):
        """Returns first found index not containing value (or list of values).
        """

        def _find(x):
            try:
                for arg in self.all:
                    if x not in arg:
                        return self.all.index(arg)
            except ValueError:
                return None

        if is_collection(x):
            for item in x:
                found = _find(item)
                if found:
                    return found
            return None
        else:
            return _find(x)

    def start_with(self, x):
        """Returns all arguments beginning with given string (or list thereof).
        """

        _args = []

        for arg in self.all:
            if is_collection(x):
                for _x in x:
                    if arg.startswith(x):
                        _args.append(arg)
                        break
            else:
                if arg.startswith(x):
                    _args.append(arg)

        return Args(_args, no_argv=True)

    def contains_at(self, x, index):
        """Tests if given [list of] string is at given index."""

        try:
            if is_collection(x):
                for _x in x:
                    if (_x in self.all[index]) or (_x == self.all[index]):
                        return True
                    else:
                        return False
            else:
                return (x in self.all[index])

        except IndexError:
            return False

    def has(self, x):
        """Returns true if argument exists at given index.
           Accepts: integer.
        """

        try:
            self.all[x]
            return True
        except IndexError:
            return False

    def value_after(self, x):
        """Returns value of argument after given found argument
        (or list thereof).
        """

        try:
            try:
                i = self.all.index(x)
            except ValueError:
                return None

            return self.all[i + 1]

        except IndexError:
            return None

    @property
    def grouped(self):
        """Extracts --flag groups from argument list.
           Returns {format: Args, ...}
        """

        collection = OrderedDict(_=Args(no_argv=True))

        _current_group = None

        for arg in self.all:
            if arg.startswith('-'):
                _current_group = arg
                collection[arg] = Args(no_argv=True)
            else:
                if _current_group:
                    collection[_current_group]._args.append(arg)
                else:
                    collection['_']._args.append(arg)

        return collection

    @property
    def last(self):
        """Returns last argument."""

        try:
            return self.all[-1]
        except IndexError:
            return None

    @property
    def all(self):
        """Returns all arguments."""

        return self._args

    def all_with(self, x):
        """Returns all arguments containing given string (or list thereof)"""

        _args = []

        for arg in self.all:
            if is_collection(x):
                for _x in x:
                    if _x in arg:
                        _args.append(arg)
                        break
            else:
                if x in arg:
                    _args.append(arg)

        return Args(_args, no_argv=True)

    def all_without(self, x):
        """Returns all arguments not containing given string (or list thereof).
        """

        _args = []

        for arg in self.all:
            if is_collection(x):
                for _x in x:
                    if _x not in arg:
                        _args.append(arg)
                        break
            else:
                if x not in arg:
                    _args.append(arg)

        return Args(_args, no_argv=True)

    @property
    def flags(self):
        """Returns Arg object including only flagged arguments."""

        return self.start_with('-')

    @property
    def not_flags(self):
        """Returns Arg object excluding flagged arguments."""

        return self.all_without('-')

    @property
    def files(self, absolute=False):
        """Returns an expanded list of all valid paths that were passed in."""

        _paths = []

        for arg in self.all:
            for path in expand_path(arg):
                if os.path.exists(path):
                    if absolute:
                        _paths.append(os.path.abspath(path))
                    else:
                        _paths.append(path)

        return _paths

    @property
    def not_files(self):
        """Returns a list of all arguments that aren't files/globs."""

        _args = []

        for arg in self.all:
            if not len(expand_path(arg)):
                if not os.path.exists(arg):
                    _args.append(arg)

        return Args(_args, no_argv=True)

    @property
    def copy(self):
        """Returns a copy of Args object for temporary manipulation."""

        return Args(self.all)


def expand_path(path):
    """Expands directories and globs in given path."""

    paths = []
    path = os.path.expanduser(path)
    path = os.path.expandvars(path)

    if os.path.isdir(path):

        for (dir, dirs, files) in os.walk(path):
            for file in files:
                paths.append(os.path.join(dir, file))
    else:
        paths.extend(glob(path))

    return paths


def is_collection(obj):
    """Tests if an object is a collection. Strings don't count."""

    if isinstance(obj, basestring):
        return False

    return hasattr(obj, '__getitem__')


class Writer(object):
    """WriterUtilized by context managers."""

    shared = dict(indent_level=0, indent_strings=[])

    def __init__(self, indent=0, quote='', indent_char=' '):
        self.indent = indent
        self.indent_char = indent_char
        self.indent_quote = quote
        if self.indent > 0:
            self.indent_string = ''.join((
                str(quote),
                (self.indent_char * (indent - len(self.indent_quote)))
            ))
        else:
            self.indent_string = ''.join((
                ('\x08' * (-1 * (indent - len(self.indent_quote)))),
                str(quote))
            )

        if len(self.indent_string):
            self.shared['indent_strings'].append(self.indent_string)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.shared['indent_strings'].pop()

    def __call__(self, s, newline=True, stream=STDOUT):
        if newline:
            s = tsplit(s, NEWLINES)
            s = map(str, s)
            indent = ''.join(self.shared['indent_strings'])

            s = (str('\n' + indent)).join(s)

        _str = ''.join((
            ''.join(self.shared['indent_strings']),
            str(s),
            '\n' if newline else ''
        ))
        stream(_str)


def puts(s='', newline=True, stream=STDOUT):
    """Prints given string to stdout via Writer interface."""
    Writer()(s, newline, stream=stream)


def indent(indent=4, quote=''):
    """Indentation context manager."""
    return Writer(indent=indent, quote=quote)


def tsplit(string, delimiters):
    """Behaves str.split but supports tuples of delimiters."""

    delimiters = tuple(delimiters)
    stack = [string, ]

    for delimiter in delimiters:
        for i, substring in enumerate(stack):
            substack = substring.split(delimiter)
            stack.pop(i)
            for j, _substring in enumerate(substack):
                stack.insert(i + j, _substring)

    return stack


def min_width(string, cols, padding=' '):
    """Returns given string with right padding."""

    stack = tsplit(str(string), NEWLINES)

    for i, substring in enumerate(stack):
        _sub = substring.ljust((cols + 0), padding)
        stack[i] = _sub

    return '\n'.join(stack)


TRUE_CHOICES = ('y', 'yes')
FALSE_CHOICES = ('n', 'no')


def process_value(value, empty=False, type=str, default=None, allowed=None,
        true_choices=TRUE_CHOICES, false_choices=FALSE_CHOICES):
    """Process prompted value.

    :param str value: The value to process.
    :param bool empty: Allow empty value.
    :param type type: The expected type.
    :param mixed default: The default value.
    :param tuple allowed: The allowed values.
    :param tuple true_choices: The accpeted values for True.
    :param tuple false_choices: The accepted values for False.
    """
    if allowed is not None and value not in allowed:
        raise Exception('Invalid input')

    if type is bool:
        if value in true_choices:
            return True
        if value in false_choices:
            return False
    if value in ('', '\n'):
        if default is not None:
            return default
        if empty:
            return None
        raise Exception('Invalid input')

    return type(value)


def prompt(message, empty=False, hidden=False, type=str, default=None,
        allowed=None, true_choices=TRUE_CHOICES, false_choices=FALSE_CHOICES,
        max_attempt=3, confirm=False):
    """Prompt user for value.

    :param str message: The prompt message.
    :param bool empty: Allow empty value.
    :param bool hidden: Hide user input.
    :param type type: The expected type.
    :param mixed default: The default value.
    :param tuple allowed: The allowed values.
    :param tuple true_choices: The accpeted values for True.
    :param tuple false_choices: The accepted values for False.
    :param int max_attempt: How many times the user is prompted back in case
        of invalid input.
    :param bool confirm: Enforce confirmation.
    """
    from manager import Error

    if allowed is not None and empty:
        allowed = allowed + ('', '\n')

    if type is bool:
        allowed = true_choices + false_choices

    if allowed is not None:
        message = "%s [%s]" % (message, ", ".join(allowed))

    if default is not None:
        message = "%s (default: %s) " % (message, default)

    handler = raw_input
    if hidden:
        handler = getpass.getpass

    attempt = 0

    while attempt < max_attempt:
        try:
            value = process_value(
                handler("%s : " % message),
                empty=empty,
                type=type,
                default=default,
                allowed=allowed,
                true_choices=true_choices,
                false_choices=false_choices,
            )
            break
        except:
            attempt = attempt + 1

            if attempt == max_attempt:
                raise Error('Invalid input')

    if confirm:
        confirmation = prompt("%s (again)" % message, empty=empty,
            hidden=hidden, type=type, default=default, allowed=allowed,
            true_choices=true_choices, false_choices=false_choices,
            max_attempt=max_attempt)

        if value != confirmation:
            raise Error('Values do not match')

    return value


class Colored(object):
    def __init__(self, color, string):
        self.color = color
        self.string = string

    def __len__(self):
        return len(self.string)

    def __str__(self):
        if sys.stdout.isatty():
            return "%s%s\033[0m" % (self.color, self.string)
        return self.string

    def __eq__(self, other):
        return self.string == other


def blue(string):
    return Colored('\033[94m', string)


def green(string):
    return Colored('\033[92m', string)


def red(string):
    return Colored('\033[91m', string)



# -*- coding: utf-8 -*-
import argparse
import collections
import functools
import sys
import re
import os
import inspect

from manager import cli


class Error(Exception):
    pass


def puts(r):
    stdout = sys.stdout.write
    type_ = type(r)
    if type_ == list:
        return [puts(i) for i in r]
    elif type_ == dict:
        for key in r:
            puts(cli.blue(cli.min_width(key, 25) + str(r[key])))
        return
    elif type_ == Error:
        return puts(cli.red(str(r)))
    elif type_ == bool:
        if r:
            return puts(cli.green('OK'))
        return puts(cli.red('FAILED'))
    elif r is not None:
        return cli.puts(str(r).rstrip('\n'), stream=stdout)


class Command(object):
    name = None
    namespace = None
    description = 'no description'
    run = None
    capture_all = False

    def __init__(self, **kwargs):
        for key in kwargs:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
            else:
                raise Exception('Invalid keyword argument `%s`' % key)

        self.args = []

        if self.name is None:
            self.name = re.sub(
                '(.)([A-Z]{1})', r'\1_\2',
                self.__class__.__name__
            ).lower()

        if not self.capture_all:
            self.inspect()

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def inspect(self):
        self.arg_names, varargs, keywords, defaults = inspect.getargspec(
            self.run)
        if hasattr(self.run, 'im_self') or hasattr(self.run, '__self__'):
            del self.arg_names[0]  # Removes `self` arg for class method
        if defaults is not None:
            kwargs = dict(zip(
                *[reversed(l) for l in (self.arg_names, defaults)]
            ))
        else:
            kwargs = []
        for arg_name in self.arg_names:
            type_ = type(kwargs[arg_name]) if arg_name in kwargs else None
            if type_ == type(None):
                type_ = None

            default = kwargs[arg_name] if arg_name in kwargs else None

            flag = None
            if type_ == bool and default is True:
                flag = 'no-%s' % arg_name

            arg = Arg(
                arg_name,
                flag = flag,
                default=default,
                type=type_,
                required=not arg_name in kwargs,
            )
            self.add_argument(arg)

    def add_argument(self, arg):
        if self.has_argument(arg.name):
            raise Exception('Arg %s already exists' % arg.name)
        self.args.append(arg)

    def get_argument(self, name):
        if not self.has_argument(name):
            raise Exception('Arg %s does not exist' % name)
        position = self.arg_names.index(name)
        return self.args[position], position

    def has_argument(self, name):
        return name in [arg.name for arg in self.args]

    def run(self, *args, **kwargs):
        raise NotImplementedError

    def parse(self, args):
        try:
            if self.capture_all:
                args, kwargs = [args], {}
            else:
                parsed_args = self.parser.parse_args(args)
                kwargs = dict(parsed_args._get_kwargs())
                args = []
                position = 0
                for arg_name in self.arg_names:
                    arg = self.args[position]
                    if not isinstance(arg, PromptedArg) and arg.required:
                        args.append(getattr(parsed_args, arg_name))
                        del kwargs[arg_name]
                    if isinstance(arg, PromptedArg):
                        args.append(arg.prompt())
                    position = position + 1
            r = self(*args, **kwargs)
            failed = r is False
        except Error as e:
            r = e
            failed = True
        puts(r)
        if failed:
            sys.exit(1)

    @property
    def parser(self):
        parser = argparse.ArgumentParser(description=self.description)
        for arg in self.args:
            if not isinstance(arg, PromptedArg):
                parser.add_argument(*arg.flags, **arg.kwargs)

        return parser

    @property
    def path(self):
        return self.name if self.namespace is None else '%s.%s' % \
            (self.namespace, self.name)


class Manager(object):
    def __init__(self, base_command=Command, envs=False):
        self.base_command = base_command
        self.commands = {}
        self.env_vars = collections.defaultdict(dict)
        if envs:
            self.command(self.envs)

    @property
    def Command(self):
        manager = self

        class BoundMeta(type):
            def __new__(meta, name, bases, dict_):
                new = type.__new__(meta, name, bases, dict_)
                if name != 'BoundCommand':
                    manager.add_command(new())
                return new

        return BoundMeta('BoundCommand', (self.base_command, ), {})

    def add_command(self, command):
        self.commands[command.path] = command

    def arg(self, name, shortcut=None, **kwargs):
        def wrapper(command):
            def wrapped(**kwargs):
                if command.has_argument(name):
                    arg, position = command.get_argument(name)
                    if shortcut is not None:
                        arg.shortcut = shortcut
                    arg._kwargs.update(**kwargs)
                    return command
                command.add_argument(Arg(name, shortcut, **kwargs))
                return command
            return wrapped(**kwargs)

        return wrapper

    def prompt(self, name, message=None, **kwargs):
        def wrapper(command):
            def wrapped(**kwargs):
                arg, position = command.get_argument(name)
                command.args[position] = PromptedArg(name, arg, message,
                    **kwargs)
                return command
            return wrapped(**kwargs)

        return wrapper

    def merge(self, manager, namespace=None):
        for command_name in manager.commands:
            command = manager.commands[command_name]
            if namespace is not None:
                command.namespace = namespace
            self.add_command(command)

    def command(self, *args, **kwargs):
        def register(fn):
            def wrapped(**kwargs):
                if not 'name' in kwargs:
                    kwargs['name'] = fn.__name__
                if not 'description' in kwargs and fn.__doc__:
                    kwargs['description'] = fn.__doc__
                command = self.Command(run=fn, **kwargs)
                self.add_command(command)
                return command
            return wrapped(**kwargs)

        if len(args) == 1 and callable(args[0]):
            fn = args[0]
            return register(fn)
        else:
            return register

    def update_env(self):
        path = os.path.join(os.getcwd(), '.env')
        if os.path.isfile(path):
            env = self.parse_env(open(path).read())
            for key in env:
                os.environ[key] = env[key]

    def parse_env(self, content):
        def strip_quotes(string):
            for quote in "'", '"':
                if string.startswith(quote) and string.endswith(quote):
                    return string.strip(quote)
            return string

        regexp = re.compile('^([A-Za-z_0-9]+)=(.*)$', re.MULTILINE)
        founds = re.findall(regexp, content)
        return {key: strip_quotes(value) for key, value in founds}

    @property
    def parser(self):
        parser = argparse.ArgumentParser(
            usage='%(prog)s [<namespace>.]<command> [<args>]')
        parser.add_argument('command', help='the command to run')
        return parser

    def usage(self):
        def format_line(command, w):
            return "%s%s" % (
                cli.min_width(command.name, w), command.description
            )

        self.parser.print_help()
        if len(self.commands) > 0:
            puts('\navailable commands:')
            with cli.indent(2):
                namespace = None
                for command_path in sorted(
                        self.commands,
                        key=lambda c: '%s%s' % (c.count('.'), c)
                ):
                    command = self.commands[command_path]
                    if command.namespace is not None:
                        if command.namespace != namespace:
                            puts(cli.red('\n[%s]' % command.namespace))
                        with cli.indent(2):
                            puts(format_line(command, 23))
                    else:
                        puts(format_line(command, 25))
                    namespace = command.namespace

    def main(self, args=None):
        args = cli.Args(args)
        if len(args) == 0 or args[0] in ('-h', '--help'):
            return self.usage()

        command = args.get(0)
        try:
            command = self.commands[command]
        except KeyError:
            puts(cli.red('Invalid command `%s`\n' % command))
            return self.usage()
        self.update_env()
        command.parse(args.all[1:])

    def env(self, key, value=None):
        """Decorator to register an ENV variable needed for a method.

        For optional env variables, set a default value using <value>.

        All env vars will be made available as key word arguments.

        @manager.env('SOME_ARG')
        def your_method(some_arg):
            ...
        """
        key = key.lower()

        def decorator(f):
            self.env_vars[f.__name__][key] = value

            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                if kwargs.get(key, None) is None:
                    if key.upper() in os.environ:
                        kwargs[key] = os.environ[key.upper()]
                    elif value:
                        kwargs[key] = value
                    else:
                        raise KeyError('Please set ENV var %s.' % key.upper())
                return f(*args, **kwargs)
            return wrapper
        return decorator

    def envs(self):
        """List required and optional environment variables."""
        if not self.env_vars:
            puts('No ENV variables have been registered.')
            puts('To register an ENV variable, use the @env(key, value) ')
            puts('method decorator of your manager object.')
            return

        puts('Registered ENV vars per method.\n')
        for func_name in self.env_vars:
            puts('%s:' % func_name)
            for var, default in self.env_vars[func_name].items():
                default = '(%s)' % default if default is not None else ''
                puts('\t%s%s' % (cli.min_width(var.upper(), 30), default))
            puts('')


class Arg(object):
    defaults = {
        'help': 'no description',
        'required': False,
        'type': None,
    }

    def __init__(self, name, flag=None, shortcut=None, **kwargs):
        self.name = name
        self.flag = flag if flag is not None else name
        self.shortcut = shortcut
        self._kwargs = self.defaults.copy()
        self._kwargs.update(kwargs)

    def __getattr__(self, key):
        if not key in self._kwargs:
            raise AttributeError
        return self._kwargs[key]

    @property
    def flags(self):
        flags = [self.flag]
        if not self.required:
            flags = ['--%s' % self.flag]
            if self.shortcut is not None:
                flags.append('-%s' % self.shortcut)
        return flags

    @property
    def kwargs(self):
        dict_ = self._kwargs.copy()

        if self.required:
            del dict_['required']
        else:
            dict_['dest'] = self.name

            if self.type == bool:
                if self.default:
                    dict_['action'] = 'store_false'
                else:
                    dict_['action'] = 'store_true'
                del dict_['type']
        return dict_


class PromptedArg(Arg):
    def __init__(self, name, arg, message=None, **kwargs):
        self.name = name
        self.message = message if message is not None else name
        self._kwargs = {
            'empty': not arg.required,
            'type': str if arg.type is None else arg.type,
            'default': arg.default,
        }
        self._kwargs.update(kwargs)

    @property
    def kwargs(self):
        return self._kwargs

    def prompt(self):
        return cli.prompt(self.message, **self.kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


setup(
    name='manage.py',
    version='0.2.3',
    description='Human friendly CLI builder',
    long_description='',
    author='jean-philippe serafin',
    author_email='jps@birdback.com',
    url='https://github.com/Birdback/manage.py',
    packages=find_packages(),
    install_requires=[],
    entry_points={
        'console_scripts': [
            'manage = manager.main:main',
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)
